﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Threading;
using System.Diagnostics;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using Clarion.Framework.Extensions;
using WorldServerLibrary.Model;

namespace ClarionSimulation { 

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// Public enum that represents all possibilities of agent actions
    public enum ClarionAgentActionType { 
    
        DO_NOTHING,
        ROTATE,
		GO_AHEAD
    }

    public class ClarionAgent { 
    
        #region Constants
        /// Constant that represents the Visual Sensor
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// Constant that represents that there is at least one wall ahead
        private String DIMENSION_WALL_AHEAD = "WallAhead";
        private String DIMENSION_WALL_FARAWAY = "WallFarAway";

		// Wall avoidance minimum distance 
        private double DIST_MIN = 61.0;		
		
		//private String GO_AHEAD = "GoalGoAhead";
		//private String ROTATE = "GoalAvoidWall";
        #endregion

        #region Properties
        #region Simulation
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        public double MaxNumberOfCognitiveCycles = -1;

        /// Current cognitive cycle number
        private double CurrentCognitiveCycle = 0;

        /// Time between cognitive cycle in miliseconds
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// A thread Class that will handle the simulation process
        private Thread runThread;
		#endregion Simulation

        #region Agent
        /// The agent 
        private Clarion.Framework.Agent CurrentAgent;

        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// For each cognitive cycle, this event will be called when the agent selects one action
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;
		#endregion Agent

        #region Perception Input
        /// Perception input to indicates a wall ahead
        private DimensionValuePair InputWallAhead;
        private DimensionValuePair InputWallFarAway;		
		#endregion Perception Input

        #region Action Output
        /// Output action that makes the agent to rotate clockwise (to avoid wall)
        private ExternalActionChunk OutputRotateClockwise;

        /// Output action that makes the agent go ahead
        private ExternalActionChunk OutputGoAhead;
		#endregion Action Output

		#region Goal Chunk
        /// Output action that makes the agent go ahead
        private GoalChunk GoalAvoidWall;
        private GoalChunk GoalGoAhead;
		#endregion Goal Chunk
        #endregion Properties

        #region Constructor
        public ClarionAgent() {         
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);
            InputWallFarAway = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_FARAWAY);

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE.ToString());
            OutputRotateClockwise.Add (World.NewDimensionValuePair("type", ClarionAgentActionType.ROTATE.ToString()));
            OutputGoAhead = World.NewExternalActionChunk(ClarionAgentActionType.GO_AHEAD.ToString());
            OutputGoAhead.Add (World.NewDimensionValuePair("type", ClarionAgentActionType.GO_AHEAD.ToString()));

            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }
		#endregion Constuctor

        #region Public Methods
        /// Run the Simulation in World Server 3d Environment
        public void Run() {                
			Console.WriteLine ("Running ...\n");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive) {             
   			    Console.WriteLine ("Setup agent infrastructure  ...\n");
				SetupAgentInfraStructure();
                if (OnNewVisualSensorialInformation != null) {                 
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// Abort the current Simulation
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent) {
            Console.WriteLine ("Aborting ...\n");
            if (runThread != null && runThread.IsAlive) {             
                runThread.Abort();
            }
            if (CurrentAgent != null && deleteAgent) {             
                CurrentAgent.Die();
            }
        }
		#endregion Public Methods

        #region Setup Agent Methods
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        private void SetupAgentInfraStructure() {        
            // Setup the ACS Subsystem
            //SetupACS();                    
            //
			//Setup the MCS and MS Subsystems
            //Console.WriteLine ("Setup MS ....... \n");
            SetupMS();                    
        }
			
        #region Deficit Change
		// Similar ao exemplo no HelloWorldFull
        private double agent_DeficitChange(ActivationCollection si, Drive target) {
			
            var currentGoal = ((SensoryInformation)si).AffiliatedAgent.CurrentGoal;
            if (currentGoal != null) {
                if ((currentGoal == GoalAvoidWall && target is AvoidingPhysicalDangerDrive) ||
                    (currentGoal == GoalGoAhead && target is CuriosityDrive))
                    target.Parameters.DEFICIT_CHANGE_RATE = .999;
                else
                    target.Parameters.DEFICIT_CHANGE_RATE = 1.001;
            }
            return target.Deficit * target.Parameters.DEFICIT_CHANGE_RATE;
        }
        #endregion Deficit Change

        #region MS
        private void SetupMS() {                    
            // Goal Chunk and add inputs and output to QBPNet
            //Console.WriteLine ("First step ... MS ... \n");			
		    GoalAvoidWall = World.NewGoalChunk("GoalAvoidWall");	
			GoalGoAhead = World.NewGoalChunk("GoalGoAhead");
			SimplifiedQBPNetwork net = AgentInitializer.InitializeImplicitDecisionNetwork(CurrentAgent, SimplifiedQBPNetwork.Factory);
            net.Input.Add(InputWallAhead); 
            net.Input.Add(GoalAvoidWall); 
            net.Input.Add(InputWallFarAway); 			
            net.Input.Add(GoalGoAhead); 			
            net.Output.Add(OutputRotateClockwise); 			
            net.Output.Add(OutputGoAhead); 			
  
            // Learning rate, agent commit with QBPNet and rule refinament for action rule   
            net.Parameters.LEARNING_RATE = 1;   
            CurrentAgent.Commit(net); 
            RefineableActionRule.GlobalParameters.INFORMATION_GAIN_OPTION = RefineableActionRule.IGOptions.PERFECT;
			
			GoalStructureUpdateActionChunk setAvoidWall = World.NewGoalStructureUpdateActionChunk();
			setAvoidWall.Add(GoalStructure.RecognizedActions.SET_RESET, GoalAvoidWall);
			GoalStructureUpdateActionChunk setGoAhead = World.NewGoalStructureUpdateActionChunk();
			setGoAhead.Add(GoalStructure.RecognizedActions.SET_RESET, GoalGoAhead);
			
			
            // MS Module: Drives and Drive Equation: physical danger drive (bis drive). Also, drive equation. 
            AvoidingPhysicalDangerDrive wallDrive = AgentInitializer.InitializeDrive(CurrentAgent, AvoidingPhysicalDangerDrive.Factory, .7, (DeficitChangeProcessor)agent_DeficitChange); 
            DriveEquation wallEquation = AgentInitializer.InitializeDriveComponent(wallDrive, DriveEquation.Factory);
            wallDrive.Commit(wallEquation);
            CurrentAgent.Commit(wallDrive);
			
            // MS Module: Drives and Drive Equation: autonomy drive (bas drive). Also, drive equation. 
            //AutonomyDrive walkDrive = AgentInitializer.InitializeDrive(CurrentAgent, AutonomyDrive.Factory, .5); 
            CuriosityDrive walkDrive = AgentInitializer.InitializeDrive(CurrentAgent, CuriosityDrive.Factory, .5, (DeficitChangeProcessor)agent_DeficitChange); 
            DriveEquation walkEquation = AgentInitializer.InitializeDriveComponent(walkDrive, DriveEquation.Factory);
            walkDrive.Commit(walkEquation);
            CurrentAgent.Commit(walkDrive);

			
            // MCS Module: Goal selection module and Goal Selection: combine goal strengths         
            GoalSelectionModule gsm = AgentInitializer.InitializeMetaCognitiveModule(CurrentAgent, GoalSelectionModule.Factory);
            GoalSelectionEquation gse = AgentInitializer.InitializeMetaCognitiveDecisionNetwork(gsm, GoalSelectionEquation.Factory);
			
            // Make goal recommendations for the goal structure based  
            gse.Input.Add(wallDrive.GetDriveStrength());    
            gse.Input.Add(walkDrive.GetDriveStrength());    

            // Goal Selection Equation: provide information to GoalSelectionModule
            GoalStructureUpdateActionChunk goalActWall = World.NewGoalStructureUpdateActionChunk();
            GoalStructureUpdateActionChunk goalActGo = World.NewGoalStructureUpdateActionChunk();

            // update the goals in MS subsystem
            goalActWall.Add(GoalStructure.RecognizedActions.SET_RESET, GoalAvoidWall); 
            goalActGo.Add(GoalStructure.RecognizedActions.SET_RESET, GoalGoAhead); 

            //Add output, set relevance and commit
            gse.Output.Add(goalActWall); 
            gse.Output.Add(goalActGo); 			
            gse.SetRelevance(goalActWall, wallDrive, 1); 
            gse.SetRelevance(goalActGo, wallDrive, 0); 
            gse.SetRelevance(goalActGo, walkDrive, 1); 
            gse.SetRelevance(goalActWall, walkDrive, 0); 
			gsm.Commit(gse);		
			CurrentAgent.Commit(gsm);			

            CurrentAgent.MS.Parameters.CURRENT_GOAL_ACTIVATION_OPTION = MotivationalSubsystem.CurrentGoalActivationOptions.FULL;			
        }
        #endregion MS

        #region ACS
        /// Setup the ACS subsystem
        private void SetupACS() {
            // Create Colission Wall Rule
            SupportCalculator avoidCollisionWallSupportCalculator = FixedRuleDelegateToAvoidColissionWall;
            FixedRule ruleAvoidCollisionWall = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotateClockwise, avoidCollisionWallSupportCalculator);

            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleAvoidCollisionWall);

            // Create Colission To Go Ahead
            SupportCalculator goAheadSupportCalculator = FixedRuleDelegateToGoAhead;
            FixedRule ruleGoAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoAhead, goAheadSupportCalculator);
            
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoAhead);

            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }
        #endregion ACS

        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation) {        
            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);

            // Detect if we have a wall ahead
            Boolean wallAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= DIST_MIN)).Any();
            double wallAheadActivationValue = wallAhead ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
            double wallFarAwayActivationValue = wallAhead ? CurrentAgent.Parameters.MIN_ACTIVATION : CurrentAgent.Parameters.MAX_ACTIVATION;
            si.Add(InputWallAhead, wallAheadActivationValue);
			si.Add(InputWallFarAway, wallFarAwayActivationValue);

			//Console.WriteLine("Avoiding physical danger drive: " + si + " \n");
            si[typeof(AvoidingPhysicalDangerDrive), AvoidingPhysicalDangerDrive.MetaInfoReservations.STIMULUS] = wallAhead ? 1 : 0;
			//Console.WriteLine("Autonomy drive: " + si + " \n");
            //si[typeof(AutonomyDrive), AutonomyDrive.MetaInfoReservations.STIMULUS] = wallAhead ? 0 : 1;
            si[typeof(CuriosityDrive), CuriosityDrive.MetaInfoReservations.STIMULUS] = wallAhead ? 0 : 1;

            return si;
        }
        #endregion

        #region Delegate Methods
        #region Fixed Rules
        private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target) {        
            // See partial match threshold to verify what are the rules available for action selection
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }

        private double FixedRuleDelegateToGoAhead(ActivationCollection currentInput, Rule target) {        
            // Here we will make the logic to go ahead
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION))) ? 1.0 : 0.0;
        }
		#endregion Fixed Rules

        #region Run Thread Method
        private void RunThread(object obj) {         
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !\n");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles) {            
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
				//Console.WriteLine("Sensory information: make perception ...... \n");
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

				// Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null) { 
				  if (CurrentAgent.CurrentGoal == GoalAvoidWall) { 	
                      if (si[InputWallAhead] == CurrentAgent.Parameters.MAX_ACTIVATION && actionType == ClarionAgentActionType.ROTATE) {
						//Console.WriteLine("Avoiding wall ............................ \n");
						Console.WriteLine("Wall ahead .. avoiding physical danger drive: ... " + si + "Action type: " + actionType + " ...... \n");
                        //si[typeof(AvoidingPhysicalDangerDrive), AvoidingPhysicalDangerDrive.MetaInfoReservations.STIMULUS] = 1;
                        CurrentAgent.ReceiveFeedback(si, 1.0);   
                        OnNewExternalActionSelected(actionType); 
                      } 
                      else {
                        //OnNewExternalActionSelected(actionType);  
						//Console.WriteLine("Wall ahead .. avoiding phiysical danger drive: .... " + si + " ....... \n");
                        CurrentAgent.ReceiveFeedback(si, 0.0); 
                      }	
					}
					else { 		
					  if (CurrentAgent.CurrentGoal == GoalGoAhead)  { 
                         if (si[InputWallFarAway] == CurrentAgent.Parameters.MAX_ACTIVATION && actionType == ClarionAgentActionType.GO_AHEAD) {
					       //Console.WriteLine("Avoiding wall ............................ \n");
						   Console.WriteLine("Wall far away .. goal go ahead - feedback 1.0 ... " + si + "Action type: " + actionType + " ...... \n");
                           CurrentAgent.ReceiveFeedback(si, 1.0);   
                           OnNewExternalActionSelected(actionType); 
                         } 
                         else {
                           //OnNewExternalActionSelected(actionType);  
						   Console.WriteLine("Wall far away .. feedback 0.0 .... " + si + " ....... \n");
                           CurrentAgent.ReceiveFeedback(si, 0.0);
                         }
					  }		
				   }
				}
                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0) {                
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }
		#endregion Run Thread Method
		#endregion Delegate Methods
    }
}
